/* COPYRIGHT NOTICE
 *
 * This program comes from "The X Toolkit Cookbook" by Paul E. Kimball
 * c. 1995 Prentice Hall.
 *
 *  HDR1: Example 6-1 - appres.c - Application resources/resource converters
 *  
 *  HDR2: Chapter 6, page 124
 *
 * This software is provided free of charge and may be freely distributed
 * in its original form, so long as this notice appears unchanged in each 
 * copy.  This software is supplied "as-is" and is not covered by warrantee
 * of any kind, express or implied.  
 *
 * This and other examples are available by anonymous ftp from
 * ftp.netcom.com, directory "pub/pk/pkimball/cookbook".
 *
 *  1/15/95	Paul Kimball - pkimball@netcom.com
 */

/***  Include Required Header Files    ******************************/

#include <X11/Intrinsic.h>       /* The Intrinsics                  */
#include <X11/StringDefs.h>      /* Intrinsics resource symbols     */
#include <Xm/Xm.h>               /* Motif symbols                   */
#include <Xm/BulletinB.h>        /* XmRowColumn widget              */
#include <Xm/PushB.h>            /* XmPushButton widget             */

/***  Declare Global Variables  *************************************/

Widget  toplevel, manager,       /* Widget tree                     */
        exit_button;
int     clicks_sofar = 0;        /* click counter                   */
void    ClickCallback();         /* A callback routine              */
Boolean CvtStringToSize();       /* Type converter String to Size   */

/*** Declare application resources  *********************************/

/*** Application data type "Size" ***/

#define XtRSize     "Size"
typedef enum {AppSMALL=1, AppMEDIUM, AppLARGE, AppXLARGE} Size;

/*** Structure to hold application resources  ***/

typedef struct {       
   int      clicks;              /* Button clicks req'd to exit     */
   Boolean  paint;               /* If True, repaint on first click */
   Pixel    color;               /* Color to paint button           */
   Size     size;                /* Relative size of pushbutton     */
 } Appres;
 
static Appres appres;            /* Declare instance of structure   */

/***  Resource list  ***/

static XtResource appres_list[] = {
  { "clicks",  "Clicks", XtRInt, sizeof(int),
     XtOffsetOf(Appres, clicks), XtRImmediate, (XtPointer)1 },
  { "paint", "Paint", XtRBoolean, sizeof(Boolean),
     XtOffsetOf(Appres, paint), XtRImmediate, (XtPointer)False },
  { "paintColor", "PaintColor", XtRPixel, sizeof(Pixel), 
     XtOffsetOf(Appres, color), XtRString, (XtPointer)XtDefaultBackground },
  { "size", "Size", XtRSize, sizeof(Size),
     XtOffsetOf(Appres, size), XtRImmediate, (XtPointer)AppMEDIUM }
};

/*** Define command-line options    *********************************/


/*  Switch      Resource affected          How to derive value      */

static XrmOptionDescRec options_list[] = {
    {"-clicks",  ".clicks",                  XrmoptionSepArg, NULL  },
    {"-paint",   ".paint",                   XrmoptionNoArg,  "true"},
    {"-color",   ".paintColor",              XrmoptionSepArg, NULL  },
    {"-size",    ".size",                    XrmoptionSepArg, NULL  },
    {"-msg",     "*exitButton.labelString",  XrmoptionSepArg, NULL  },
    {"-bigfont", "*fontList",                XrmoptionNoArg,
                 "-*-*-Bold-R-Normal--*-240-*-*-*-*-ISO8859-1"}
};

/*** Define fallback resources   ************************************/

static String fallback_resources[] = {
    "*paint:        true",
    ".paintColor:   red",
    ".clicks:       1",
    ".size:         MEDIUM",
    "*exitButton.labelString: Push to Exit",
    "*fontList:     -*-Helvetica-*-R-*--*-120-*-*-*-*-ISO8859-1",
    NULL };

/***  Main program        *******************************************/

main(argc,argv)
int argc;
char **argv;
{

 XtAppContext  context;             /* Application context          */

 toplevel = XtAppInitialize( &context, "Appres",
                             NULL, 0, &argc, argv, NULL, NULL, 0);

/*** Register private converter and fetch application resources *****/

 XtAppSetTypeConverter(      /* Register converter, String-to-Size  */
     context,                /* Application context                 */
     XtRString,              /* "From" type handled by converter    */
     XtRSize,                /* "To" type handled by converter      */
     CvtStringToSize,        /* The converter                       */
     NULL, 0,                /* Additional args - unused here       */
     XtCacheAll,             /* Save the value in case needed later */
     NULL);

 XtGetApplicationResources( toplevel, &appres, 
     appres_list, XtNumber(appres_list), NULL, 0);

/***  Create widget tree, bind callbacks  ***************************/

 manager = XtVaCreateManagedWidget( "manager",
     xmBulletinBoardWidgetClass, toplevel, NULL );

 exit_button = XtVaCreateManagedWidget( "exitButton",
     xmPushButtonWidgetClass, manager, 
     XmNwidth, 100*appres.size,            /* Size determined by    */
     XmNheight, 20*appres.size,            /* application resources */
     NULL );

 XtAddCallback( exit_button, XmNactivateCallback, ClickCallback, NULL );

 XtRealizeWidget(toplevel);                /* Realize all           */
 XtAppMainLoop(context);                   /* Start Event Loop      */
}

/*** Callback Routines  *********************************************/

void ClickCallback(w, client_data, call_data) 
Widget w;
XtPointer client_data;
XtPointer call_data;
{
 if (appres.paint) 
     XtVaSetValues(exit_button,            /* Repaint the button    */
     XmNbackground,  appres.color,         /* Pixel value from      */
     NULL );                               /* resource converter    */

 if (++clicks_sofar >= appres.clicks)      /* clicked enough        */
     exit(1);
}


/*** A Custom Resource Converter ************************************/

Boolean 
CvtStringToSize(display, args, num_args, from, to, convert_data)
Display  *display;                      /* Display pointer          */
XrmValue  *args;                        /* Additional run-time args */
Cardinal  *num_args;                    /* Number of add. arguments */
XrmValue  *from, *to;                   /* input and output values  */
XtPointer *convert_data;                /* Other data               */
{ 

 String       str = (String)from->addr;
 static Size  result = 0;
 Boolean      convert_failed = False;

/* Perform the conversion */

 if      (strcmp(str, "SMALL") == 0)  result = AppSMALL;
 else if (strcmp(str, "MEDIUM") == 0) result = AppMEDIUM;
 else if (strcmp(str, "LARGE") == 0)  result = AppLARGE;
 else if (strcmp(str, "XLARGE") == 0) result = AppXLARGE;
 else convert_failed = True;

 if (convert_failed) {                  /* Not recognized           */
   XtDisplayStringConversionWarning(    /* Print a warning          */
     display,                           /* Associated display       */
     str,                               /* String not converted     */
     XtRSize);                          /* Failed "to" type         */
 }

/* Try returning value if conversion succeeded */

 else {
   if( to->addr == NULL )               /* to.addr is NULL          */
     to->addr = (XPointer)&result;      /* Send address to caller   */
   else {
     if( to->size < sizeof(Size) )      /* No room; convert fails   */
       convert_failed = True;
     else *(Size*)(to->addr) = result;  /* Enough room; copy value  */
   }
   to->size = sizeof(Size);             /* Actual or required size  */
 }
 return (!convert_failed);              /* Conversion result        */
}
