/* COPYRIGHT NOTICE
 *
 * This program comes from "The X Toolkit Cookbook" by Paul E. Kimball
 * c. 1995 Prentice Hall.
 *
 *  HDR1: Example 12-2 graphics.c - A simple graphics editor
 *  
 *  HDR2: Chapter 12, page 383
 *
 * This software is provided free of charge and may be freely distributed
 * in its original form, so long as this notice appears unchanged in each 
 * copy.  This software is supplied "as-is" and is not covered by warrantee
 * of any kind, express or implied.  
 *
 * This and other examples are available by anonymous ftp from
 * ftp.netcom.com, directory "pub/pk/pkimball/cookbook".
 *
 *  1/15/95	Paul Kimball - pkimball@netcom.com
 */

/***  Include Required Header Files       ****************************/

#include <X11/Intrinsic.h>              /* The Intrinsics            */
#include <X11/StringDefs.h>             /* Resources names           */
#include <X11/Shell.h>                  /* Shell widgets             */
#include <Xm/Form.h>                    /* XmForm                    */
#include <Xm/RowColumn.h>               /* XmRowColumn               */
#include <Xm/PushB.h>                   /* XmPushButton              */
#include <Xm/ToggleB.h>                 /* XmToggleButton            */
#include <Xm/Text.h>                    /* XmText                    */
#include <Xm/ScrolledW.h>               /* XmScrolledWindow          */
#include <Xm/DrawingA.h>                /* XmDrawingArea             */

/***  Declare Global Variables            ****************************/

XtAppContext context;                   /* Application context       */
Widget toplevel, manager;               /* Widget tree               */
Widget menu, scroll_win, canvas;
Widget text_area, clear_btn, exit_btn, 
       obj_menu, line_btn, text_btn,
       style_menu, wide_btn, thin_btn;

/***  Display list for saving application graphics  ******************/

typedef enum { LINE, RECT, TEXT } ObjectType;
typedef enum { WIDE, THIN } LineWidth;

#define MAXOBJS 1000                    /* Max size of display list  */

int n, num_objects = 0;                 /* Array counters            */
struct object_record {                  /* Object record             */
  ObjectType type;                      /* type of object            */
  Position x1, y1;                      /* First endpoint            */
  Position x2, y2;                      /* Second endpoint           */
  char *text;                           /* Pointer to char string    */
  GC drawing_gc;                        /* Line width, style, font   */
  }  display_list[MAXOBJS];             /* Array of records          */

/***  Variables for drawing and maintaining refresh state  ***********/

ObjectType current_obj = LINE;          /* Type of object to drawn   */
String text_input;                      /* Input from text area      */
Position start_x, start_y;              /* Initial pointer position  */ 
Position last_x, last_y;                /* Current pointer position  */
Boolean rubberbanding = False;          /* True = rubberband in prog */
Dimension canvas_width, canvas_height;  /* Wid, Hgt of drawing area  */
Pixmap backing_pixmap;                  /* Pixmap refresh buffer     */
Boolean pixmap_buffer = False;          /* True = backing pixmap used*/
Region expose_region, null_region;      /* Exposure boundaries       */
GC copy_gc, expose_gc, clear_gc,        /* GCs for drawing ops       */
   wide_gc, thin_gc, current_gc, xor_gc;

/*****  Declare application Resources ********************************/

typedef struct {                        /* Declare structure type    */
 Pixel drawing_color;                   /* Pixel(color) for graphics */
 Font  drawing_font;                    /* Font for drawn text       */
 } Appres;
Appres appres;                          /* Struct to hold values     */
static XtResource appres_list[] = {
  { "drawingColor", "Foreground", XtRPixel, sizeof(Pixel),
     XtOffsetOf(Appres, drawing_color),
     XtRString, "XtDefaultForeground"},
  {"drawingFont", "Font", XtRFont, sizeof(Font),
     XtOffsetOf(Appres, drawing_font),
     XtRString, "XtDefaultFont"},
};

/*****  Forward declare callback routines   **************************/

void ExitCallback();       /* Callback - exits                       */
void SetlineCallback();    /* Callback - sets line width             */
void SetobjCallback();     /* Callback - sets item type to draw next */
void ExposeCallback();     /* Callback - Refreshes canvas if exposed */
void ClearCallback();      /* Callback - Clears graphics             */
void StartRubber();        /* Callback - Starts rubberband drawing   */
void Rubberband();         /* Ev. Handler - draws in rubberband mode */
void EndRubber();          /* Callback - Draws object in final form  */
void InitGraphics();       /* Function - Inits GCs, refresh mode     */
void StoreObject();        /* Function - Stores item in disp. list   */
void DrawObject();         /* Function - Draws item from disp. list  */
void XorObject();          /* Function - Draws cur. item in xor mode */

/*****  Start of main()   ********************************************/

main(argc,argv)
unsigned int argc;
char **argv;
{

/*****  Init. Toolkit, Appl. Context, Application Shell  *************/

 toplevel = XtAppInitialize(&context, "Graphics",
     NULL, 0, &argc, argv, NULL, NULL, 0);

 XtGetApplicationResources(toplevel,              /* appl. resources */
     &appres, appres_list, 
     XtNumber(appres_list), NULL, 0);

/*****  Create widget tree, bind callbacks  **************************/

 manager = XtCreateManagedWidget("mainw",      /* Create main window */
     xmFormWidgetClass, toplevel, NULL, 0);    /*  and drawing area  */

 scroll_win = XtCreateManagedWidget("scrollWin", 
     xmScrolledWindowWidgetClass, manager, NULL, 0 );

 canvas = XtCreateManagedWidget("canvas",
     xmDrawingAreaWidgetClass, scroll_win, NULL, 0 );

 /* Trap exposures to refresh graphics */
 XtAddCallback(canvas, XmNexposeCallback, ExposeCallback, NULL );

 /* Trap pointer button press and release   */
 XtAddCallback(canvas, XmNinputCallback, StartRubber, NULL );
 XtAddCallback(canvas, XmNinputCallback, EndRubber, NULL );

 /* Trap motion events to track cursor */
 XtAddEventHandler(canvas, Button1MotionMask, False, Rubberband, NULL);

 menu = XtCreateManagedWidget("menu",             /* Create menu of  */
     xmRowColumnWidgetClass, manager, NULL, 0);   /*  drawing ops    */

 obj_menu = XtCreateManagedWidget("objectMenu",   /* Menu of toggle  */ 
     xmRowColumnWidgetClass, menu, NULL, 0);      /*  buttons        */

 line_btn = XtCreateManagedWidget ("lineBtn",     /* "Line" button   */
     xmToggleButtonWidgetClass, obj_menu, NULL, 0);
 XtAddCallback(line_btn, XmNvalueChangedCallback,
     SetobjCallback, (XtPointer)LINE );

 text_btn = XtCreateManagedWidget ("textBtn",     /* "Text" button   */
     xmToggleButtonWidgetClass, obj_menu, NULL, 0);
 XtAddCallback(text_btn, XmNvalueChangedCallback,
     SetobjCallback, (XtPointer)TEXT );

 text_area = XtCreateManagedWidget("textArea",    /* Text input area */
     xmTextWidgetClass, menu, NULL, 0);

 style_menu = XtCreateManagedWidget("styleMenu",  /* Menu of Toggle  */
     xmRowColumnWidgetClass, menu, NULL, 0);      /*  buttons        */

 wide_btn = XtCreateManagedWidget ("wideBtn",     /* "Wide" button   */
     xmToggleButtonWidgetClass, style_menu, NULL, 0);
 XtAddCallback(wide_btn, XmNvalueChangedCallback,
     SetlineCallback, (XtPointer)WIDE );

 thin_btn = XtCreateManagedWidget ("thinBtn",     /* "Thin" button   */
     xmToggleButtonWidgetClass, style_menu, NULL, 0);
 XtAddCallback(thin_btn, XmNvalueChangedCallback,
     SetlineCallback, (XtPointer)THIN );

 clear_btn = XtCreateManagedWidget ("clearBtn",   /* "Clear" button  */
     xmPushButtonWidgetClass, menu, NULL, 0);
 XtAddCallback(clear_btn, XmNactivateCallback, ClearCallback, NULL );

 exit_btn = XtCreateManagedWidget ("exitBtn",     /* "Exit" button   */
     xmPushButtonWidgetClass, menu, NULL, 0);
 XtAddCallback(exit_btn, XmNactivateCallback, ExitCallback, NULL ); 

/***  Set resources that can't be set in resource file   ***/

 XtVaSetValues( scroll_win, XmNleftAttachment, XmATTACH_WIDGET,
     XmNleftWidget, menu, NULL);

 XtRealizeWidget(toplevel);             /**  Realize everybody  ******/
 InitGraphics();                        /**  Init GCs and graphics  **/

 XtAppMainLoop(context);                /**  Start Main Loop   *******/
}

/*** Application Callback routines  **********************************/

void                                    /* Start rubberband line     */
StartRubber(w, client_data, call_data)
Widget w;                               /* Widget ID of drawing area */
XtPointer client_data, call_data;       /* Call data holds event     */
{
 /* Check call_data for button press event */
 XmAnyCallbackStruct *cbs = (XmAnyCallbackStruct*)call_data;
 if ((cbs->event->type == ButtonPress) 
         && (cbs->event->xbutton.button == Button1) ) {
   rubberbanding = True;                /* Start rubberband drawing  */
   start_x = cbs->event->xbutton.x;     /* Save first endpoint       */
   start_y = cbs->event->xbutton.y;
   last_x = start_x;                    /* Get latest pointer pos.   */
   last_y = start_y; 
   if(current_obj == TEXT)              /* Get text input value      */
     XtVaGetValues( text_area, XmNvalue, &text_input, NULL);
   XorObject();                         /* Draw first XORed object   */
 }
}

/*********************************************************************/

void                                    /* Rubberband objects while  */
Rubberband(w, client_data, event, dis)  /*  pointer is moved         */
Widget w;                               /* Widget ID of drawing area */
XtPointer client_data;                  /* unused                    */
XEvent *event;                          /* MotionNotify event        */
Boolean *dis;                           /* unused                    */
{
 if ((event->type == MotionNotify) && (rubberbanding)) {
   XorObject();                         /* XOR to erase last object  */
   last_x = event->xmotion.x;           /* Get latest pointer coords */
   last_y = event->xmotion.y;
   XorObject();                         /* XOR to draw new object    */
 } 
}

/*********************************************************************/

void                                    /* Finish drawing object     */
EndRubber(w, client_data, call_data)    /*  when button released     */
Widget w;                               /* unused                    */
XtPointer client_data, call_data;       /* unused                    */
{
 XmAnyCallbackStruct *cbs = (XmAnyCallbackStruct*)call_data;
 if ((cbs->event->type == ButtonRelease)  /* Finish this object      */ 
     && (cbs->event->xbutton.button == 1) ) { 
   rubberbanding = False;               /* Stop rubberbanding        */
   XorObject();                         /* XOR erase last object     */
   last_x = cbs->event->xbutton.x;      /* Get final pointer         */
   last_y = cbs->event->xbutton.y;      /*  coordinates              */
   StoreObject();                       /* Put in display list       */
 }
}

/*********************************************************************/

void                                    /* Store object in list and  */
StoreObject()                           /*  draw in final rendition  */
{       
 if (num_objects == MAXOBJS) {          /* Attempt to exceed array   */
   XBell(XtDisplay(canvas), 0);         /* Ring bell to signal       */  
   return;
 }
 else {                                 /* Store in display list     */
   n = num_objects;
   display_list[n].type = current_obj;
   display_list[n].drawing_gc = current_gc;
   display_list[n].x1 = last_x;         /* Current pointer position  */
   display_list[n].y1 = last_y;
   switch ((int)current_obj) {
     case LINE:                         /* If this is a line -       */
       display_list[n].x2 = start_x;    /*  second endpoint of line  */
       display_list[n].y2 = start_y;    /*  is original start point  */
       break;
     case TEXT:                         /* If text, copy input to    */
       display_list[n].text =           /*  display list record      */
         XtNewString(text_input);       /* Allocate a new string     */
       break;
   }
   DrawObject(n);                       /* Draw object on screen     */
   num_objects++ ;                      /* Increment num_objects     */
   if (pixmap_buffer)                   /* If pixmap buffer on, copy */
     XCopyArea(XtDisplay(canvas),       /*  to visible window        */
        backing_pixmap, XtWindow(canvas),
        copy_gc, 0, 0, canvas_width, canvas_height, 0, 0);
 }
}

/*********************************************************************/

void                                    /* Draw object in            */
XorObject()                             /*  XOR mode                 */
{
 switch ((int)current_obj){
  case LINE:
    XDrawLine(XtDisplay(canvas), XtWindow(canvas),
        xor_gc, last_x, last_y, start_x, start_y );
    break;
  case TEXT:
    XDrawString(XtDisplay(canvas), XtWindow(canvas),
        xor_gc, last_x, last_y, text_input, strlen(text_input));
    break;
 }
}

/*********************************************************************/

void                                    /* Draw object from          */
DrawObject(n)                           /*  display list             */
int n;                                  /* Index of object           */
{
 Drawable d;                            /* Draw into buffer          */
 if(pixmap_buffer) d = backing_pixmap;  /* pixmap, if exists         */
 else d = XtWindow(canvas);             /* else draw to window       */
  switch ((int)display_list[n].type) {
   case LINE:                           /* Draw a line               */
    XDrawLine(XtDisplay(canvas), d,
       display_list[n].drawing_gc,
       display_list[n].x1, display_list[n].y1,
       display_list[n].x2, display_list[n].y2 );
    break;
   case TEXT:                           /* Draw text                 */
    XDrawString(XtDisplay(canvas), d,
        display_list[n].drawing_gc,
        display_list[n].x1, display_list[n].y1,
        display_list[n].text, strlen(display_list[n].text) );
    break;
  }
}

/*********************************************************************/

void                                    /* Fix up graphics           */
ExposeCallback(w, client_data, call_data)
Widget w;                               /* Widget ID of canvas       */
XtPointer client_data, call_data;       /* Drawing area data         */
{
 XmAnyCallbackStruct *cbs = (XmAnyCallbackStruct*)call_data;

 /** Add exposure area to clipping region **/ 
 if (pixmap_buffer) XtAddExposureToRegion(cbs->event, expose_region);

 if (cbs->event->xexpose.count > 0) {   /* Don't draw unless this    */
   return;                              /*  is last Expose event     */
 }
 else {                                 /* Last exposure             */

   if (pixmap_buffer){
     /* Copy pixmap using region to mask copied area   */
     XSetRegion(XtDisplay(w), expose_gc, expose_region);
     XCopyArea(XtDisplay(w),
       backing_pixmap, XtWindow(w),
       expose_gc, 0, 0,canvas_width, canvas_height,0,0);
     XIntersectRegion(null_region, expose_region, expose_region);
   }
   else {                               /* No buffer exists, so      */
     for ( n=0; n<num_objects; n++ )  { /*  redraw all objects -     */
       DrawObject(n);                   /* Inefficient but works     */
     }
   }
 }
}

/*********************************************************************/

void                                    /* Clear drawing area        */
ClearCallback(w, client_data, call_data)/*  backing store            */
Widget w;                               /* unused                    */
XtPointer client_data, call_data;       /* unused                    */
{
 XClearWindow( XtDisplay(canvas),       /* Clear drawing area        */ 
      XtWindow(canvas));                /*  to default background    */
 if(pixmap_buffer)                      /* Clear pixmap, if used     */
   XFillRectangle(XtDisplay(canvas),    /* Pixmap has no default     */ 
      backing_pixmap, clear_gc,         /*  background, so fill      */
   0, 0, canvas_width, canvas_height);  /*  with canvas bg color     */
 num_objects = 0;                       /* Reset object counter      */
}

/*********************************************************************/

void                                        /* Set GC used to        */
SetlineCallback(w, client_data, call_data)  /*  draw lines           */
Widget w;                                   /* unused                */
XtPointer client_data;                      /* Line width            */
XtPointer call_data;                        /* unused                */
{
 switch((int)client_data){
   case WIDE: current_gc = wide_gc; break;  /* Use wide GC           */
   case THIN: current_gc = thin_gc; break;  /* Use thin GC           */
 }
}

/*********************************************************************/

void                                        /* Set type of object    */
SetobjCallback(w, client_data, call_data)
Widget w;                                   /* unused                */
XtPointer client_data;                      /* Type of object        */
XtPointer call_data;                        /* unused                */
{
  switch((int)client_data){
    case LINE: current_obj = LINE; break;
    case TEXT: current_obj = TEXT; break;
  }
}

/*********************************************************************/

void                                    /* Initialize drawing GCs    */
InitGraphics()                          /*  and refresh strategy     */
{

 Pixel                fg, bg;           /* Canvas fg and bg pixels   */
 XGCValues            gcvals;           /* GC atb data structure     */
 XtGCMask             gcvals_mask;      /* Mask of GC attributes     */
 unsigned long        mask;             /* Mask of window attributes */
 XSetWindowAttributes attributes;       /* Window atb data structure */
 int                  depth;            /* Depth of canvas in planes */

 XtVaGetValues( canvas, 
     XtNforeground, &fg,                    /* Get canvas fg &           */
     XtNbackground, &bg,                    /*  bg pixel vals            */
     NULL);       

 copy_gc = XCreateGC(XtDisplay(canvas),           /* This GC is not      */
     XtWindow(canvas), NULL, &gcvals);            /*  clipped            */

 expose_gc = XCreateGC(XtDisplay(canvas),         /* This GC clips to    */
     XtWindow(canvas), NULL, &gcvals);            /*  exposure region    */

 gcvals.foreground = bg;                          /* GC to draw in       */
 gcvals_mask = GCForeground;                      /*  win bg color       */
 clear_gc = XCreateGC(XtDisplay(canvas),          /*  Use it to          */
     XtWindow(canvas), gcvals_mask, &gcvals);     /*  clear pixmap       */

 gcvals.font = appres.drawing_font;               /* Create a GC for     */
 gcvals.foreground = appres.drawing_color;        /* normal              */
 gcvals.function = GXcopy;                        /* thin lines          */
 gcvals.line_width = 1; 
 gcvals_mask = GCForeground | GCFunction | GCFont | GCLineWidth;
 thin_gc = XCreateGC(XtDisplay(canvas),
     XtWindow(canvas), gcvals_mask, &gcvals);

 gcvals.line_width = 3;                           /* wide lines          */
 gcvals_mask = GCForeground | GCFunction | GCFont | GCLineWidth;
 wide_gc = XCreateGC(XtDisplay(canvas),
     XtWindow(canvas), gcvals_mask, &gcvals);

 gcvals.foreground = bg ^ appres.drawing_color;   /* Create a GC for */
 gcvals.function = GXxor;                         /*  XOR lines      */
 gcvals_mask = GCForeground | GCFont | GCFunction;
 xor_gc = XCreateGC(XtDisplay(canvas),
     XtWindow(canvas), gcvals_mask, &gcvals);

/*** Refresh strategy one - use server backing store if available ****/

 if ( XDoesBackingStore(XtScreen(canvas)) ) { /* Server has backing  */
   attributes.backing_store = WhenMapped;     /*  store, so set the  */
   mask = CWBackingStore;                     /*  window attribute   */
   XChangeWindowAttributes(XtDisplay(canvas), /*  for drawing area   */
       XtWindow(canvas), mask, &attributes);
   printf("Using backing store \n");
 }

/*** Refresh strategy two - use off-screen pixmap as refresh buffer **/

 else {
   XtVaGetValues(canvas,                        /* Create pixmap with  */
       XtNwidth, &canvas_width,                 /*  width, height,     */
       XtNheight, &canvas_height,               /*  depth as drawing   */
       XtNdepth, &depth,                        /*  widget             */
       NULL);

   if ( backing_pixmap = XCreatePixmap( XtDisplay(canvas),
       XtWindow(canvas), canvas_width, canvas_height, depth) ) {

       pixmap_buffer = True;                    /* False if failed     */

       XFillRectangle(XtDisplay(canvas),        /* Clear pixmap        */
           backing_pixmap, clear_gc, 0, 0,      /*  to background      */ 
           canvas_width, canvas_height);        /*  color              */

       expose_region = XCreateRegion();         /* Create clip region  */
       null_region = XCreateRegion();           /* Create null region  */

       printf("Using buffer pixmap \n");
   }
   else {
       /* Fallback if all else fails */
       printf("Resorting to explicit redraw on exposures \n");
   }
 }
 num_objects = 0;                             /* Init display list   */ 
 current_gc = thin_gc;                        /* Set default gc      */
}

/*********************************************************************/

void                                          /* Exit application    */
ExitCallback(w, client_data, call_data)
Widget w;                                     /* unused              */
XtPointer client_data, call_data;             /* unused              */
{
 exit(1);                                     /* exit                */
}

